"""
This is a module that contains functions responsible for mutating
the trajectory file for alanine scanning in gmx_MMPBSA. It must be
included with gmx_MMPBSA to insure proper functioning of alanine
scanning.
"""

# ##############################################################################
#                           GPLv3 LICENSE INFO                                 #
#                                                                              #
#  Copyright (C) 2020  Mario S. Valdes-Tresanco and Mario E. Valdes-Tresanco   #
#  Copyright (C) 2014  Jason Swails, Bill Miller III, and Dwight McGee         #
#                                                                              #
#   Project: https://github.com/Valdes-Tresanco-MS/gmx_MMPBSA                  #
#                                                                              #
#   This program is free software; you can redistribute it and/or modify it    #
#  under the terms of the GNU General Public License version 3 as published    #
#  by the Free Software Foundation.                                            #
#                                                                              #
#  This program is distributed in the hope that it will be useful, but         #
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
#  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    #
#  for more details.                                                           #
# ##############################################################################

from GMXMMPBSA.exceptions import MutateError, MutantResError

#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

def _getCoords(line, coordsperline, coordsize):
    """ Returns the coordinates of a line in a mdcrd file """
    holder = []
    location = 0
    for i in range(coordsperline):
        try:
            tmp = float(line[location:location+coordsize])
            holder.append(tmp)
        except:
            pass
        location += coordsize

    return holder

#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

def _scaledistance(coords, dist):
    """ Scales the distance between 2 3-D cartesian coordinates to the specified
        distance
    """
    from math import sqrt

    if len(coords) != 6:
        raise MutateError('_scaledistance requires x,y,z coords for 2 atoms')

    coords[3] -= coords[0]  # set first 3 coordinates as origin
    coords[4] -= coords[1]
    coords[5] -= coords[2]

    actualdist = sqrt(coords[3]*coords[3] + coords[4]*coords[4] +
                      coords[5]*coords[5])

    scalefactor = dist / actualdist # determine scale factor

    coords[3] *= scalefactor  # scale original coordinates
    coords[4] *= scalefactor
    coords[5] *= scalefactor

    coords[3] += coords[0] # move back to original place
    coords[4] += coords[1]
    coords[5] += coords[2]

    return coords  # return the coordinates

#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

def _getnumatms(resname):
    """ Returns the number of atoms in a given Amino acid residue """
    if resname in 'GLY':
        return 7
    if resname in ('ALA', 'CYM', 'CYX'):
        return 10
    if resname in ('CYS', 'SER'):
        return 11
    if resname in ('ASP'):
        return 12
    if resname in ('ASN', 'PRO', 'THR'):
        return 14
    if resname in ('GLU'):
        return 15
    if resname in ('GLH', 'VAL'):
        return 16
    if resname in ('GLN', 'HID', 'HIE', 'MET'):
        return 17
    if resname in ('HIP'):
        return 18
    if resname in ('ILE', 'LEU'):
        return 19
    if resname in ('PHE'):
        return 20
    if resname in ('LYN', 'TYR'):
        return 21
    if resname in ('LYS'):
        return 22
    if resname in ('ARG', 'TRP'):
        return 24

    raise MutateError(('Unrecognized residue! Add %s to _getnumatms(resname) in alamdcrd.py and reinstall gmx_MMPBSA'
                       % resname))

#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

def _ressymbol(resname):
    """ Return 1-letter symbol of give amino acid """

    if resname == 'ALA':
        return 'A'
    elif resname == 'ARG':
        return 'R'
    elif resname == 'ASN':
        return 'N'
    elif resname in ['ASP','ASH']:
        return 'D'
    elif resname in ['CYS','CYX','CYM']:
        return 'C'
    elif resname in ['GLU','GLH']:
        return 'E'
    elif resname == 'GLN':
        return 'Q'
    elif resname == 'GLY':
        return 'G'
    elif resname in ['HIP','HID','HIE']:
        return 'H'
    elif resname == 'ILE':
        return 'I'
    elif resname == 'LEU':
        return 'L'
    elif resname in ['LYN','LYS']:
        return 'K'
    elif resname == 'MET':
        return 'M'
    elif resname == 'PHE':
        return 'F'
    elif resname == 'PRO':
        return 'P'
    elif resname == 'SER':
        return 'S'
    elif resname == 'THR':
        return 'T'
    elif resname == 'TRP':
        return 'W'
    elif resname == 'TYR':
        return 'Y'
    elif resname == 'VAL':
        return 'V'
    else:
        return resname

#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class MutantMdcrd(object):
    """ Class for an alanine-mutated amber trajectory file.
        ASCII only (no netcdf)
    """

    #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def __init__(self, trajname, prm1, prm2):
        self.traj = trajname
        self.orig_prm = prm1
        self.new_prm = prm2
        self.mutres = self.FindMutantResidue()
        self.hasbox = bool(prm1.ptr('ifbox'))

    #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def __str__(self):
        return '%s%d%s' % (_ressymbol(
            self.orig_prm.parm_data['RESIDUE_LABEL'][self.mutres-1]),
                           self.mutres, _ressymbol('ALA'))

    #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def FindMutantResidue(self):
        """ Finds which residue is the alanine mutant in a pair of prmtop files
        """
        origres = self.orig_prm.parm_data['RESIDUE_LABEL']
        newres = self.new_prm.parm_data['RESIDUE_LABEL']
        diffs = 0
        mutres = -1

        if len(origres) != len(newres):
            raise MutateError(('Mutant prmtop (%s) has a different number of ' +
                               'residues than the original (%s)!') %
                              (self.new_prm.prm_name, self.orig_prm.prm_name))

        for i in range(len(origres)):
            if origres[i] != newres[i]:
                diffs += 1
                if newres[i] != 'ALA':
                    raise MutantResError('Mutant residue %s is %s but must be ALA!' %
                                         (i+1, newres[i]))
                mutres = i + 1

        if diffs == 0:
            raise MutateError('Your mutant prmtop (%s) has the same sequence as the original!' % self.new_prm.prm_name)
        elif diffs > 1:
            raise MutateError('Your mutant prmtop (%s) can only have one mutation!' % self.new_prm.prm_name)

        return mutres

    #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def MutateTraj(self, newname):
        """ Mutates a given mdcrd file based on 2 prmtops """

        mutres = self.mutres

        orig_resname = self.orig_prm.parm_data['RESIDUE_LABEL'][mutres-1]
        resstart = self.orig_prm.parm_data['RESIDUE_POINTER'][mutres-1]
        try:
            nextresstart = self.orig_prm.parm_data['RESIDUE_POINTER'][mutres]
        except IndexError:
            nextresstart = self.orig_prm.ptr('natom')
        number_atoms_mut = self.new_prm.ptr('natom')
        if self.orig_prm.ptr('ifbox'):
            number_atoms_mut += 1

        coordsperline = 10 # number of coordinates in each line
        coordsize = 8      # how large coordinates are in characters
        counter = 0
        coords_done = 0
        coords_tomutate = []
        temp_holder = []
        new_coords = []

        # location is 0 before modified coordinates, 1 during modified
        # coordinates, and 2 after modified coordinates
        location = 0

        if orig_resname == 'GLY':
            raise MutateError('You are trying to mutate GLY to ALA! ' +
                              'Not currently supported.')

        new_mdcrd = open(newname, 'w')
        mdcrd = open(self.traj, 'r')

        for line in mdcrd:
            counter += 1
            # First line is always a comment
            if counter == 1:
                new_mdcrd.write('%-80s' % (line.strip() + ' and mutated by gmx_MMPBSA for alanine scanning'))
                continue

            if location == 0 and coords_done <= max(resstart * 3 - 4, 0) and \
                    coords_done + coordsperline > max(resstart * 3 - 4, 0):
                location = 1
                words = _getCoords(line, coordsperline, coordsize)
                for i in range(coordsperline):
                    if coords_done <= resstart * 3 - 4:
                        if coords_done % coordsperline == 0:
                            new_mdcrd.write('\n')
                        new_mdcrd.write('%8.3f' % words[i])
                        coords_done += 1
                    else:
                        coords_tomutate.append(words[i])
                continue

            elif location == 1:
                words = _getCoords(line, coordsperline, coordsize)
                if coordsperline + len(coords_tomutate) >= \
                        3 * (nextresstart - resstart):
                    location = 2
                    for i in range(coordsperline):
                        if len(coords_tomutate) < 3 * (nextresstart - resstart):
                            coords_tomutate.append(words[i])
                        else:
                            temp_holder.append(words[i])

                    new_coords = self._mutate(orig_resname, coords_tomutate)
                    for i in range(len(new_coords)):
                        if coords_done % coordsperline == 0:
                            new_mdcrd.write('\n')
                        new_mdcrd.write('%8.3f' % new_coords[i])
                        coords_done += 1
                    if len(temp_holder) != 0:
                        for i in range(len(temp_holder)):
                            if coords_done % coordsperline == 0:
                                new_mdcrd.write('\n')
                            new_mdcrd.write('%8.3f' % temp_holder[i])
                            coords_done += 1
                    coords_tomutate = []
                    temp_holder = []
                else:
                    for i in range(coordsperline):
                        coords_tomutate.append(words[i])

                continue

            elif location == 2:
                words = _getCoords(line, coordsperline, coordsize)
                for i in range(len(words)):
                    if coords_done % coordsperline == 0:
                        if not self.hasbox or coords_done < number_atoms_mut * 3 - 3:
                            new_mdcrd.write('\n')
                    new_mdcrd.write('%8.3f' % words[i])
                    coords_done += 1
                    if self.hasbox and coords_done == number_atoms_mut * 3 - 3:
                        new_mdcrd.write('\n')

                if coords_done == number_atoms_mut * 3:
                    coords_done = 0
                    location = 0
                continue
            else:
                if coords_done % coordsperline == 0:
                    new_mdcrd.write('\n')
                new_mdcrd.write(line[:len(line)-1])
                coords_done = coords_done + coordsperline
                continue

        new_mdcrd.write('\n')
        mdcrd.close()
        new_mdcrd.close()

    #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def _mutate(self, resname, coords):
        list_one = 'ARG ASH ASN ASP CYM CYS CYX GLH GLN GLU HID HIE HIP ' + \
                   'LEU LYN LYS MET PHE SER TRP TYR'
        list_two = 'ILE THR VAL'
        list_three = 'PRO'

        chdist = 1.09
        nhdist = 1.01

        coords_tosend = []
        new_coords  = []
        coords_received = []
        cterm = False

        if _getnumatms(resname) * 3 == len(coords):
            startindex = 0
            cterm = False
        elif (_getnumatms(resname) + 2) * 3 == len(coords):
            startindex = 2
            cterm = False
        elif (_getnumatms(resname) + 1) * 3 == len(coords):
            startindex = 0
            cterm = True
        else:
            raise MutateError(('Mismatch in atom # in residue %s. (%d in alamdcrd.py and %d passed in)') %
                              (resname, _getnumatms(resname), len(coords)))

        if resname in list_one:
            for i in range((7+startindex)*3):
                new_coords.append(coords[i])
            for i in range(3):
                coords_tosend.append(coords[(4+startindex)*3+i])
            for i in range(3):
                coords_tosend.append(coords[(7+startindex)*3+i])
            coords_received = _scaledistance(coords_tosend, chdist)

            for i in range(3):
                new_coords.append(coords_received[i+3])

        elif resname in list_two:
            for i in range((6+startindex)*3):
                new_coords.append(coords[i])

            for i in range(3):
                coords_tosend.append(coords[(4+startindex)*3+i])
            for i in range(3):
                coords_tosend.append(coords[(6+startindex)*3+i])
            coords_received = _scaledistance(coords_tosend, chdist)

            for i in range(3):
                new_coords.append(coords_received[i+3])

            coords_tosend = []
            coords_received = []
            for i in range(3):
                coords_tosend.append(coords[(4+startindex)*3+i])
            for i in range(3):
                coords_tosend.append(coords[(10+startindex)*3+i])
            coords_received = _scaledistance(coords_tosend, chdist)

            for i in range(3):
                new_coords.append(coords_received[3+i])

        elif resname in list_three:
            for i in range((1+startindex)*3):
                new_coords.append(coords[i])

            coords_tosend = coords[startindex*3:startindex*3 + 6]
            coords_received = _scaledistance(coords_tosend, nhdist)

            for i in range(3):
                new_coords.append(coords_received[i+3])
            for i in range(6):
                new_coords.append(coords[(10+startindex)*3+i])
            for i in range(9):
                new_coords.append(coords[(7+startindex)*3+i])

            coords_tosend = []
            coords_received = []
            for i in range(3):
                coords_tosend.append(coords[(7+startindex)*3+i])
            for i in range(3):
                coords_tosend.append(coords[(4+startindex)*3+i])
            coords_received = _scaledistance(coords_tosend, chdist)

            for i in range(3):
                new_coords.append(coords_received[i+3])

        else:
            raise MutateError("Residue %s not recognized! Can't mutate." % resname)

        if cterm:
            for i in range(9):
                new_coords.append(coords[len(coords)-9+i])
        else:
            for i in range(6):
                new_coords.append(coords[len(coords)-6+i])

        return new_coords

#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class GlyMutantMdcrd(MutantMdcrd):

    #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def __str__(self):
        return '%s%d%s' % (_ressymbol(
            self.orig_prm.parm_data['RESIDUE_LABEL'][self.mutres-1]),
                           self.mutres, _ressymbol('GLY'))

    #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def _mutate(self, resname, coords):
        list_one = 'ARG ASH ASN ASP CYM CYS CYX GLH GLN GLU HID HIE HIP ' + \
                   'LEU LYN LYS MET PHE SER TRP TYR ALA ILE THR VAL'
        list_two = 'PRO'

        chdist = 1.09
        nhdist = 1.01

        coords_tosend = []    # Coordinates to send to be scaled
        new_coords  = []      # Mutated coordinates to return
        coords_received = []  # Scaled coordinates received
        cterm = False

        if _getnumatms(resname) * 3 == len(coords):
            startindex = 0
            cterm = False
        elif (_getnumatms(resname) + 2) * 3 == len(coords):
            startindex = 2
            cterm = False
        elif (_getnumatms(resname) + 1) * 3 == len(coords):
            startindex = 0
            cterm = True
        else:
            raise MutateError(('Mismatch in atom # in residue %s. (%d in ' +
                               'alamdcrd.py and %d passed in)') % (resname,
                                                                   _getnumatms(resname), len(coords)))

        if resname in list_one:
            for i in range((4+startindex)*3):
                new_coords.append(coords[i])
            for i in range(3):
                coords_tosend.append(coords[(2+startindex)*3+i])
            for i in range(3):
                coords_tosend.append(coords[(4+startindex)*3+i])
            coords_received = _scaledistance(coords_tosend, chdist)

            for i in range(3):
                new_coords.append(coords_received[i+3])

        elif resname in list_two:
            for i in range((1+startindex)*3):
                new_coords.append(coords[i])

            coords_tosend = coords[startindex*3:startindex*3 + 6]
            coords_received = _scaledistance(coords_tosend, nhdist)

            for i in range(3):
                new_coords.append(coords_received[i+3])
            for i in range(6):
                new_coords.append(coords[(10+startindex)*3+i])

            coords_tosend = []
            coords_received = []
            for i in range(3):
                coords_tosend.append(coords[(10+startindex)*3+i])
            for i in range(3):
                coords_tosend.append(coords[(7+startindex)*3+i])
            coords_received = _scaledistance(coords_tosend, chdist)

            for i in range(3):
                new_coords.append(coords_received[i+3])

        else:
            raise MutateError("Residue %s not recognized! Can't mutate." % resname)

        if cterm:
            for i in range(9):
                new_coords.append(coords[len(coords)-9+i])
        else:
            for i in range(6):
                new_coords.append(coords[len(coords)-6+i])

        return new_coords

    #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def FindMutantResidue(self):
        """ Finds which residue is the alanine mutant in a pair of prmtop files
        """
        origres = self.orig_prm.parm_data['RESIDUE_LABEL']
        newres = self.new_prm.parm_data['RESIDUE_LABEL']
        diffs = 0
        mutres = -1

        if len(origres) != len(newres):
            raise MutateError(('Mutant prmtop (%s) has a different number of ' +
                               'residues than the original (%s)!') %
                              (self.new_prm.prm_name, self.orig_prm.prm_name))

        for i in range(len(origres)):
            if origres[i] != newres[i]:
                diffs += 1
                if newres[i] != 'GLY':
                    raise MutantResError('Mutant residue %s is %s but must be GLY!' %
                                         (i+1, newres[i]))
                mutres = i + 1

        if diffs == 0:
            raise MutateError(('Your mutant prmtop (%s) has the same sequence ' +
                               'as the original!') % (self.new_prm.prm_name,
                                                      self.orig_prm.prm_name))
        elif diffs > 1:
            raise MutateError('Your mutant prmtop (%s) can only have one mutation!'
                              % self.new_prm.prm_name)

        return mutres



"""
This module is responsible for creating the final output file for MM/PBSA
statistics printing.
"""

# ##############################################################################
#                           GPLv3 LICENSE INFO                                 #
#                                                                              #
#  Copyright (C) 2020  Mario S. Valdes-Tresanco and Mario E. Valdes-Tresanco   #
#  Copyright (C) 2014  Jason Swails, Bill Miller III, and Dwight McGee         #
#                                                                              #
#   Project: https://github.com/Valdes-Tresanco-MS/gmx_MMPBSA                  #
#                                                                              #
#   This program is free software; you can redistribute it and/or modify it    #
#  under the terms of the GNU General Public License version 3 as published    #
#  by the Free Software Foundation.                                            #
#                                                                              #
#  This program is distributed in the hope that it will be useful, but         #
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
#  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    #
#  for more details.                                                           #
# ##############################################################################

import io
from GMXMMPBSA.amber_outputs import EnergyVector
from GMXMMPBSA.exceptions import LengthError
from GMXMMPBSA import utils
from math import sqrt, ceil
from os import linesep as ls


# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

def write_stability_output(app):
    """ Writes output files based on stability calculations """
    import csv
    # Load some objects into top-level name space
    FILES = app.FILES
    INPUT = app.INPUT
    mut_str = app.mut_str

    if FILES.energyout:
        ene_csv = open(FILES.energyout, 'w')
        energyvectors = csv.writer(ene_csv, dialect='excel')

    final_output = OutputFile(FILES.output_file, 'w')
    final_output.write_date()
    final_output.add_comment('')
    final_output.write(app.input_file_text)
    final_output.print_file_info(FILES, INPUT)
    final_output.add_comment('')
    final_output.add_comment('Calculations performed using %s complex frames.' %
                             app.numframes)
    if INPUT['nmoderun']:
        final_output.add_comment('NMODE calculations performed using %s frames.' %
                                 app.numframes_nmode)

    if INPUT['pbrun']:
        if INPUT['sander_apbs']:
            final_output.add_comment('Poisson Boltzmann calculations performed ' +
                                     'using iAPBS interface to sander (sander.APBS)')
        elif INPUT['use_sander'] or INPUT['decomprun']:
            final_output.add_comment('Poisson Boltzmann calculations performed ' +
                                     'using internal PBSA solver in sander.')
        else:
            final_output.add_comment('Poisson Boltzmann calculations performed ' +
                                     'using internal PBSA solver in mmpbsa_py_energy')

    final_output.add_comment('')

    if INPUT['gbrun']:
        if INPUT['molsurf']:
            final_output.add_comment('Generalized Born ESURF calculated using '
                                     '\'molsurf\' surface areas')
        else:
            final_output.add_comment('Generalized Born ESURF calculated using '
                                     '\'LCPO\' surface areas')
        final_output.add_comment('')

    final_output.add_comment('All units are reported in kcal/mole.')
    if INPUT['nmoderun'] or INPUT['entropy']:
        if INPUT['entropy'] == 2:
            final_output.add_comment('All entropy results have units kcal/mol ' +
                                 '(Temperature is %.2f K).' % INPUT['entropy_temp'])
        else:
            final_output.add_comment('All entropy results have units kcal/mol ' +
                                     '(Temperature is %.2f K).' % INPUT['temp'])
    if INPUT['ifqnt']:
        final_output.add_comment(('QM/MM: Residues %s are treated with the ' +
                                  'Quantum Hamiltonian %s') % (INPUT['qm_residues'], INPUT['qm_theory']))
    final_output.separate()

    # Start with entropies
    if INPUT['entropy'] == 1:
        if not INPUT['mutant_only']:
            qhnorm = app.calc_types['qh']
            final_output.writeline('ENTROPY RESULTS (QUASI-HARMONIC ' +
                                   'APPROXIMATION) CALCULATED WITH PTRAJ:')
            final_output.add_section(qhnorm.print_summary())
        if INPUT['alarun']:
            qhmutant = app.calc_types['mutant']['qh']
            final_output.writeline(mut_str + ' MUTANT')
            final_output.writeline('ENTROPY RESULTS (QUASI-' +
                                   'HARMONIC APPROXIMATION) CALCULATED WITH PTRAJ:')
            final_output.add_section(qhmutant.print_summary())
        if INPUT['alarun'] and not INPUT['mutant_only']:
            final_output.add_section(('\nRESULT OF ALANINE SCANNING: (%s) DELTA ' +
                                      'DELTA S binding = %9.4f\n') % (mut_str,
                                                                      qhnorm.total_avg() - qhmutant.total_avg()))
    # end if INPUT['entropy']

    # Now print out the normal mode results
    if INPUT['nmoderun']:
        if not INPUT['mutant_only']:
            nm_norm = app.calc_types['nmode']['complex']
            final_output.write('ENTROPY RESULTS (HARMONIC APPROXIMATION) ' +
                               'CALCULATED WITH NMODE:\n')
            final_output.add_section('Complex:\n' + nm_norm.print_summary())
            # Now dump the energy vectors in CSV format
            if FILES.energyout:
                energyvectors.writerow(['NMODE entropy results'])
                nm_norm.print_vectors(energyvectors)
                energyvectors.writerow([])

        if INPUT['alarun']:
            nm_mut = app.calc_types['mutant']['nmode']['complex']
            final_output.write(mut_str + ' MUTANT\nENTROPY RESULTS (HARMONIC ' +
                               'APPROXIMATION) CALCULATED WITH NMODE:\n')
            final_output.add_section('Complex:\n' + nm_mut.print_summary())
            # Now dump the energy vectors in CSV format
            if FILES.energyout:
                energyvectors.writerow([mut_str + ' Mutant NMODE entropy results'])
                nm_mut.print_vectors(energyvectors)
                energyvectors.writerow([])

        # Now calculate the effect of alanine scanning
        if INPUT['alarun'] and not INPUT['mutant_only']:
            try:
                nm_diff_array = nm_norm.data['Total'] - nm_mut.data['Total']
                nm_davg = nm_diff_array.avg()
                nm_dstdev = nm_diff_array.stdev()
            except LengthError:
                nm_diff_array = []
                nm_davg = nm_norm.data['Total'].avg() - nm_mut.data['Total'].avg()
                nm_dstdev = sqrt(nm_norm.data['Total'].stdev() ** 2 +
                                 nm_mut.data['Total'].stdev() ** 2)
            final_output.add_section(('\nRESULT OF ALANINE SCANNING: (%s) DELTA ' +
                                      'DELTA S binding = %9.4f +/- %9.4f\n') % (mut_str,
                                                                                nm_davg, nm_dstdev))

    # end if INPUT['nmoderun']

    # Now print out the Free Energy results
    triggers = ('gbrun', 'pbrun', 'rismrun_std', 'rismrun_gf')
    outkeys = ('gb', 'pb', 'rism std', 'rism gf')
    headers = ('\nGENERALIZED BORN:\n\n', '\nPOISSON BOLTZMANN:\n\n',
               '\n3D-RISM:\n\n', '\n3D-RISM (Gauss. Fluct.):\n\n')
    for i, key in enumerate(outkeys):
        if not INPUT[triggers[i]]:
            continue
        if not INPUT['mutant_only']:
            com_norm = app.calc_types[key]['complex']
            final_output.write(headers[i])
            final_output.add_section('Complex:\n' + com_norm.print_summary())
            # Dump energy vectors to a CSV
            if FILES.energyout:
                energyvectors.writerow([headers[i].strip()])
                com_norm.print_vectors(energyvectors)
                energyvectors.writerow([])

            # Combine with the entropy(ies)
            if INPUT['entropy'] == 1:
                final_output.add_section('Using Quasi-harmonic Entropy ' +
                                         'Approximation: FREE ENERGY (G) = %9.4f\n' %
                                         (com_norm.data['TOTAL'].avg() - qhnorm.total_avg()))
            if INPUT['nmoderun']:
                if len(com_norm.data['TOTAL']) != len(nm_norm.data['Total']):
                    davg = (com_norm.data['TOTAL'].avg() -
                            nm_norm.data['Total'].avg())
                    dstdev = sqrt(com_norm.data['TOTAL'].stdev() ** 2 +
                                  nm_norm.data['Total'].stdev() ** 2)
                else:
                    diff_array = com_norm.data['TOTAL'] - nm_norm.data['Total']
                    davg = diff_array.avg()
                    dstdev = diff_array.stdev()

                final_output.add_section('Using Normal Mode Entropy Approxima' +
                                         'tion: FREE ENERGY (G) =    %9.4f +/- %7.4f\n' % (davg, dstdev))

        if INPUT['alarun']:
            com_mut = app.calc_types['mutant'][key]['complex']
            final_output.write('%s MUTANT:%s' % (mut_str, headers[i]))
            final_output.add_section('Complex:\n' + com_mut.print_summary())
            # Dump energy vectors to a CSV
            if FILES.energyout:
                energyvectors.writerow([mut_str + ' Mutant ' + headers[i]])
                com_mut.print_vectors(energyvectors)
                energyvectors.writerow([])

            # Combine with the entropy(ies)
            if INPUT['entropy'] == 1:
                final_output.add_section('Using Quasi-harmonic Entropy ' +
                                         'Approximation: DELTA G binding = %9.4f\n' %
                                         (com_mut.data['TOTAL'].avg() - qhmutant.total_avg()))
            if INPUT['nmoderun']:
                if len(com_mut.data['TOTAL']) != len(nm_mut.data['Total']):
                    davg = (com_mut.data['TOTAL'].avg() -
                            nm_mut.data['Total'].avg())
                    dstdev = sqrt(com_mut.data['TOTAL'].stdev() ** 2 +
                                  nm_mut.data['Total'].stdev() ** 2)
                else:
                    diff_array = com_mut.data['TOTAL'] - nm_mut.data['Total']
                    davg = diff_array.avg()
                    dstdev = diff_array.stdev()

                final_output.add_section('Using Normal Mode Entropy Approxima' +
                                         'tion: FREE ENERGY (G) =    %9.4f +/- %7.4f\n' % (davg, dstdev))

        if INPUT['alarun'] and not INPUT['mutant_only']:
            diff_array = com_norm.data['TOTAL'] - com_mut.data['TOTAL']
            davg = diff_array.avg()
            dstdev = diff_array.stdev()
            final_output.write(('\nRESULT OF ALANINE SCANNING: (%s) DELTA ' +
                                'G = %9.4f  +/- %9.4f\n') % (mut_str, davg, dstdev))
            if INPUT['entropy'] == 1:
                final_output.write(('\n   (quasi-harmonic entropy) (%s) DELTA ' +
                                    'G = %9.4f\n') % (mut_str,
                                                      davg + qhnorm.total_avg() - qhmutant.total_avg()))
            if INPUT['nmoderun']:
                if len(nm_diff_array) == len(diff_array):
                    total_diffs = diff_array - nm_diff_array
                    davg1 = total_diffs.avg()
                    dstdev1 = total_diffs.stdev()
                else:
                    davg1 = davg - nm_davg
                    dstdev1 = sqrt(dstdev ** 2 - nm_dstdev ** 2)
                final_output.write(('\n      (normal mode entropy) (%s) DELTA ' +
                                    'G = %9.4f +/- %9.4f\n') % (mut_str, davg1, dstdev1))
            final_output.separate()

    # end for solv in ['gbrun', 'pbrun', ...]

    if FILES.energyout: ene_csv.close()


# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

def write_binding_output(app):
    """ Writes a binding output file """
    import csv

    FILES = app.FILES
    INPUT = app.INPUT
    mut_str = app.mut_str
    prmtop_system = app.normal_system

    # Open the energy vector CSV output file if we are writing one
    if FILES.energyout:
        ene_csv = open(FILES.energyout, 'w')
        energyvectors = csv.writer(ene_csv, dialect='excel')

    # Open the file and write some initial data to it
    final_output = OutputFile(FILES.output_file, 'w')
    final_output.write_date()
    final_output.add_comment('')
    final_output.write(app.input_file_text)
    final_output.print_file_info(FILES, INPUT)
    final_output.add_comment('')
    final_output.add_comment('Receptor mask:                  "%s"' %
                             INPUT['receptor_mask'])
    final_output.add_comment('Ligand mask:                    "%s"' %
                             INPUT['ligand_mask'])
    if prmtop_system.ligand_prmtop.ptr('nres') == 1:
        final_output.add_comment('Ligand residue name is "%s"' %
                                 prmtop_system.ligand_prmtop.parm_data['RESIDUE_LABEL'][0])
    final_output.add_comment('')
    final_output.add_comment('Calculations performed using %s complex frames.' %
                             app.numframes)
    if INPUT['nmoderun']:
        final_output.add_comment('NMODE calculations performed using %s frames.' %
                                 app.numframes_nmode)
    if INPUT['entropy'] == 2:
        final_output.add_comment('Interaction Entropy calculations performed using last %s frames.' %
                                 ceil(app.numframes * (INPUT['entropy_seg']/100)))


    if INPUT['pbrun']:
        if INPUT['sander_apbs']:
            final_output.add_comment('Poisson Boltzmann calculations performed ' +
                                     'using iAPBS interface to sander (sander.APBS)')
        elif INPUT['use_sander'] or INPUT['decomprun']:
            final_output.add_comment('Poisson Boltzmann calculations performed ' +
                                     'using internal PBSA solver in sander.')
        else:
            final_output.add_comment('Poisson Boltzmann calculations performed ' +
                                     'using internal PBSA solver in mmpbsa_py_energy')
    final_output.add_comment('')

    if INPUT['gbrun']:
        if INPUT['molsurf']:
            final_output.add_comment('Generalized Born ESURF calculated using '
                                     '\'molsurf\' surface areas')
        else:
            final_output.add_comment('Generalized Born ESURF calculated using '
                                     '\'LCPO\' surface areas')
        final_output.add_comment('')

    final_output.add_comment('All units are reported in kcal/mole.')
    if INPUT['nmoderun'] or INPUT['entropy']:
        if INPUT['entropy'] == 2:
            final_output.add_comment('All entropy results have units kcal/mol ' +
                                     '(Temperature is %.2f K).' % INPUT['entropy_temp'])
        else:
            final_output.add_comment('All entropy results have units kcal/mol ' +
                                     '(Temperature is %.2f K).' % INPUT['temp'])
    if INPUT['ifqnt']:
        final_output.add_comment(('QM/MM: Residues %s are treated with the ' +
                                  'Quantum Hamiltonian %s') % (INPUT['qm_residues'], INPUT['qm_theory']))
    final_output.separate()

    # First do the entropies
    if INPUT['entropy'] == 1:
        if not INPUT['mutant_only']:
            qhnorm = app.calc_types['qh']
            final_output.writeline('ENTROPY RESULTS (QUASI-HARMONIC APPROXIMATION) CALCULATED WITH PTRAJ:')
            final_output.add_section(qhnorm.print_summary())
        if INPUT['alarun']:
            qhmutant = app.calc_types['mutant']['qh']
            final_output.writeline(mut_str + ' MUTANT')
            final_output.writeline('ENTROPY RESULTS (QUASI-HARMONIC APPROXIMATION) CALCULATED WITH PTRAJ:')
            final_output.add_section(qhmutant.print_summary())
        if INPUT['alarun'] and not INPUT['mutant_only']:
            final_output.add_section(('\nRESULT OF ALANINE SCANNING: (%s) DELTA ' +
                                      'DELTA S binding = %9.4f\n') % (mut_str,
                                                                      qhnorm.total_avg() - qhmutant.total_avg()))
    # end if INPUT['entropy']
    if INPUT['entropy'] == 2:
        if not INPUT['mutant_only']:
            ienorm = app.calc_types['ie']
            final_output.writeline('ENTROPY RESULTS (INTERACTION ENTROPY):')
            final_output.add_section(ienorm.print_summary())
        if INPUT['alarun']:
            iemutant = app.calc_types['mutant']['ie']
            final_output.writeline(mut_str + ' MUTANT')
            final_output.writeline('ENTROPY RESULTS (INTERACTION ENTROPY):')
            final_output.add_section(iemutant.print_summary())
        if INPUT['alarun'] and not INPUT['mutant_only']:
            final_output.add_section(('\nRESULT OF ALANINE SCANNING: (%s) DELTA ' +
                                      'DELTA S binding = %9.4f\n') % (mut_str, ienorm.value - iemutant.value))


    # Now print out the normal mode results
    if INPUT['nmoderun']:
        if not INPUT['mutant_only']:
            nm_sys_norm = app.calc_types['nmode']['delta']
            final_output.write('ENTROPY RESULTS (HARMONIC APPROXIMATION) ' +
                               'CALCULATED WITH NMODE:\n\n')
            final_output.add_section(nm_sys_norm.print_summary())
            # Now dump the energy vectors in CSV format
            if FILES.energyout:
                energyvectors.writerow(['NMODE entropy results'])
                nm_sys_norm.print_vectors(energyvectors)
                energyvectors.writerow([])

        if INPUT['alarun']:
            nm_sys_mut = app.calc_types['mutant']['nmode']['delta']
            final_output.write(mut_str + ' MUTANT\nENTROPY RESULTS (HARMONIC ' +
                               'APPROXIMATION) CALCULATED WITH NMODE:\n')
            final_output.add_section(nm_sys_mut.print_summary())
            # Now dump the energy vectors in CSV format
            if FILES.energyout:
                energyvectors.writerow([mut_str + ' Mutant NMODE entropy results'])
                nm_sys_mut.print_vectors(energyvectors)
                energyvectors.writerow([])

        # Now calculate the effect of alanine scanning
        if INPUT['alarun'] and not INPUT['mutant_only']:
            davg, dstdev = nm_sys_norm.diff(nm_sys_mut, 'Total', 'Total')
            final_output.add_section(('\nRESULT OF ALANINE SCANNING: (%s) DELTA ' +
                                      'DELTA S binding = %9.4f +/- %9.4f\n') % (mut_str, davg,
                                                                                dstdev))

    # end if INPUT['nmoderun']

    triggers = ('gbrun', 'pbrun', 'rismrun_std', 'rismrun_gf')
    outkeys = ('gb', 'pb', 'rism std', 'rism gf')
    headers = ('\nGENERALIZED BORN:\n\n', '\nPOISSON BOLTZMANN:\n\n',
               '\n3D-RISM:\n\n', '\n3D-RISM (Gauss. Fluct.):\n\n')
    # Now print out the Free Energy results
    for i, key in enumerate(outkeys):
        if not INPUT[triggers[i]]:
            continue

        if not INPUT['mutant_only']:
            sys_norm = app.calc_types[key]['delta']
            final_output.write(headers[i])
            final_output.add_section(sys_norm.print_summary())
            # Dump energy vectors to a CSV
            if FILES.energyout:
                energyvectors.writerow([headers[i].strip()])
                sys_norm.print_vectors(energyvectors)
                energyvectors.writerow([])

            # Combine with the entropy(ies)
            if INPUT['entropy'] == 1:
                if isinstance(sys_norm.data['DELTA TOTAL'], EnergyVector):
                    final_output.add_section('Using Quasi-harmonic Entropy ' +
                                             'Approximation: DELTA G binding = %9.4f\n' %
                                             (sys_norm.data['DELTA TOTAL'].avg() - qhnorm.total_avg()))
                else:
                    final_output.add_section('Using Quasi-harmonic Entropy ' +
                                             'Approximation: DELTA G binding = %9.4f\n' %
                                             (sys_norm.data['DELTA TOTAL'][0] - qhnorm.total_avg()))
            elif INPUT['entropy'] == 2:
                if isinstance(sys_norm.data['DELTA TOTAL'], EnergyVector):
                    final_output.add_section('Using Interaction Entropy ' +
                                             'Approximation: DELTA G binding = %9.4f\n' %
                                             (sys_norm.data['DELTA TOTAL'].avg() + ienorm.value))
                else:
                    final_output.add_section('Using Interaction Entropy ' +
                                             'Approximation: DELTA G binding = %9.4f\n' %
                                             (sys_norm.data['DELTA TOTAL'][0] + ienorm.value))
            if INPUT['nmoderun']:
                davg, dstdev = sys_norm.diff(nm_sys_norm, 'DELTA TOTAL', 'Total')
                final_output.add_section('Using Normal Mode Entropy '
                                         'Approximation: DELTA G binding =    %9.4f +/- %7.4f\n' % (davg, dstdev))

        if INPUT['alarun']:
            sys_mut = app.calc_types['mutant'][key]['delta']
            final_output.write('%s MUTANT:%s' % (mut_str, headers[i]))
            final_output.add_section(sys_mut.print_summary())
            # Dump energy vectors to a CSV
            if FILES.energyout:
                energyvectors.writerow([mut_str + ' Mutant ' + headers[i]])
                sys_mut.print_vectors(energyvectors)
                energyvectors.writerow([])

            # Combine with the entropy(ies)
            if INPUT['entropy'] == 1:
                if isinstance(sys_mut.data['DELTA TOTAL'], EnergyVector):
                    final_output.add_section('Using Quasi-harmonic Entropy ' +
                                             'Approximation: DELTA G binding = %9.4f\n' %
                                             (sys_mut.data['DELTA TOTAL'].avg() - qhmutant.total_avg()))
                else:
                    final_output.add_section('Using Quasi-harmonic Entropy ' +
                                             'Approximation: DELTA G binding = %9.4f\n' %
                                             (sys_mut.data['DELTA TOTAL'][0] - qhmutant.total_avg()))
            elif INPUT['entropy'] == 2:
                if isinstance(sys_mut.data['DELTA TOTAL'], EnergyVector):
                    final_output.add_section('Using Interaction Entropy ' +
                                             'Approximation: DELTA G binding = %9.4f\n' %
                                             (sys_mut.data['DELTA TOTAL'].avg() + iemutant.value))
                else:
                    final_output.add_section('Using Interaction Entropy ' +
                                             'Approximation: DELTA G binding = %9.4f\n' %
                                             (sys_mut.data['DELTA TOTAL'][0] + iemutant.value))
            if INPUT['nmoderun']:
                davg, dstdev = sys_mut.diff(nm_sys_mut, 'DELTA TOTAL', 'Total')
                final_output.add_section('Using Normal Mode Entropy Approxima' +
                                         'tion: DELTA G binding =    %9.4f +/- %7.4f\n' % (davg, dstdev))

        if INPUT['alarun'] and not INPUT['mutant_only']:
            davg, dstdev = sys_norm.diff(sys_mut, 'DELTA TOTAL', 'DELTA TOTAL')
            final_output.write(('\nRESULT OF ALANINE SCANNING: (%s) DELTA ' +
                                'DELTA G binding = %9.4f  +/- %9.4f\n') % (mut_str, davg, dstdev))
            if INPUT['entropy'] == 1:
                final_output.write(('\n   (quasi-harmonic entropy) (%s) DELTA ' +
                                    'DELTA G binding = %9.4f\n') % (mut_str,
                                                                    davg + qhnorm.total_avg() - qhmutant.total_avg()))
            elif INPUT['entropy'] == 2:
                final_output.write(('\n   (interaction entropy) (%s) DELTA ' +
                                    'DELTA G binding = %9.4f\n') % (mut_str, davg + ienorm.value - iemutant.value))
            if INPUT['nmoderun']:
                davg1, dstdev1 = nm_sys_norm.diff(nm_sys_mut, 'Total', 'Total')
                final_output.write(('\n      (normal mode entropy) (%s) DELTA ' +
                                    'DELTA G binding = %9.4f +/- %9.4f\n') % (mut_str, davg - davg1,
                                                                              sqrt(dstdev ** 2 + dstdev1 ** 2)))
            final_output.separate()

    # end for solv in ['gbrun', 'pbrun', ...]

    if FILES.energyout:
        ene_csv.close()


# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

def write_decomp_stability_output(FILES, INPUT, size, prmtop_system,
                                  mutant_system, mutstr, pre):
    """ Write output file for stability decomposition calculations """
    from csv import writer
    from datetime import datetime
    from GMXMMPBSA.amber_outputs import DecompOut, PairDecompOut, idecompString

    if INPUT['idecomp'] in [1, 2]:
        DecompClass = DecompOut
    else:
        DecompClass = PairDecompOut

    # Open up the files
    if INPUT['csv_format']:
        dec_out_file = OutputFile(FILES.decompout, 'wb', 0)
        decompout = writer(dec_out_file)
        decompout.writerow(['| Run on %s' % datetime.now().ctime()])
        if INPUT['gbrun']:
            decompout.writerow(['| GB non-polar solvation energies calculated '
                                'with gbsa=2'])
        decompout.writerow([idecompString[INPUT['idecomp']]])
    else:
        dec_out_file = OutputFile(FILES.decompout, 'w', 0)
        decompout = dec_out_file
        decompout.write_date()
        if INPUT['gbrun']:
            decompout.add_comment('| GB non-polar solvation energies calculated '
                                  'with gbsa=2')
        decompout.writeline(idecompString[INPUT['idecomp']])

    # Open up the energyvector CSV file
    if FILES.dec_energies: dec_energies = open(FILES.dec_energies, 'w')

    # GB first
    if INPUT['gbrun']:
        # See if we're dumping all energy terms
        csv_prefix = ''
        if FILES.dec_energies: csv_prefix = pre + 'gb'

        # See if we do normal system
        if not INPUT['mutant_only']:
            gb_com = DecompClass(pre + 'complex_gb.mdout',
                                 prmtop_system.complex_prmtop, INPUT['surften'], csv_prefix,
                                 size, INPUT['dec_verbose'])
            gb_com.fill_all_terms()
            if INPUT['csv_format']:
                decompout.writerow(['Energy Decomposition Analysis (All units ' +
                                    'kcal/mol): Generalized Born Solvent'])
                decompout.writerow([])
                gb_com.write_summary_csv(gb_com.numframes, decompout)
                decompout.writerow([])
            else:
                decompout.writeline('Energy Decomposition Analysis (All units ' +
                                    'kcal/mol): Generalized Born Solvent')
                decompout.writeline('')
                gb_com.write_summary(gb_com.numframes, decompout)
                decompout.writeline('')
            # Copy over all temporary .csv files created by gb_com to the total
            # utils.concatenate removes the temporary files after they are copied
            if FILES.dec_energies:
                # Close out the CSV files
                del gb_com.csvwriter
                dec_energies.write('Generalized Born Decomposition Energies' + ls)
                for token in gb_com.allowed_tokens:
                    utils.concatenate(dec_energies, csv_prefix + '.' + token + '.csv')
                dec_energies.write(ls)

        if INPUT['alarun']:
            gb_com = DecompClass(pre + 'mutant_complex_gb.mdout',
                                 mutant_system.complex_prmtop, INPUT['surften'], csv_prefix,
                                 size, INPUT['dec_verbose'])
            gb_com.fill_all_terms()
            if INPUT['csv_format']:
                decompout.writerow(['Energy Decomposition Analysis (All units ' +
                                    'kcal/mol): Generalized Born Solvent (%s)' % mutstr])
                decompout.writerow([])
                gb_com.write_summary_csv(gb_com.numframes, decompout)
                decompout.writerow([])
            else:
                decompout.writeline('Energy Decomposition Analysis (All units ' +
                                    'kcal/mol): Generalized Born Solvent (%s)' % mutstr)
                decompout.writeline('')
                gb_com.write_summary(gb_com.numframes, decompout)
                decompout.writeline('')
            # Copy over all temporary .csv files created by gb_com to the total
            # utils.concatenate removes the temporary files after they are copied
            if FILES.dec_energies:
                # Close out the CSV files
                del gb_com.csvwriter
                dec_energies.write('GB Decomposition Energies (%s mutant)' %
                                   mutstr + ls)
                for token in gb_com.allowed_tokens:
                    utils.concatenate(dec_energies, csv_prefix + '.' + token + '.csv')
                dec_energies.write(ls)

    # PB next
    if INPUT['pbrun']:
        # See if we're dumping all energy terms
        csv_prefix = ''
        if FILES.dec_energies: csv_prefix = pre + 'pb'

        # See if we do normal system
        if not INPUT['mutant_only']:
            pb_com = DecompClass(pre + 'complex_pb.mdout',
                                 prmtop_system.complex_prmtop, INPUT['surften'], csv_prefix,
                                 size, INPUT['dec_verbose'])
            pb_com.fill_all_terms()
            if INPUT['csv_format']:
                decompout.writerow(['Energy Decomposition Analysis (All units ' +
                                    'kcal/mol): Generalized Born Solvent'])
                decompout.writerow([])
                pb_com.write_summary_csv(pb_com.numframes, decompout)
                decompout.writerow([])
            else:
                decompout.writeline('Energy Decomposition Analysis (All units ' +
                                    'kcal/mol): Poisson Boltzmann Solvent')
                decompout.writeline('')
                pb_com.write_summary(pb_com.numframes, decompout)
                decompout.writeline('')
            # Dump the energy vectors if requested
            if FILES.dec_energies:
                # Close out the CSV files
                del pb_com.csvwriter
                dec_energies.write('Poisson Boltzmann Decomposition Energies' + ls)
                for token in pb_com.allowed_tokens:
                    utils.concatenate(dec_energies, csv_prefix + '.' + token + '.csv')
                dec_energies.write(ls)
        if INPUT['alarun']:
            pb_com = DecompClass(pre + 'mutant_complex_pb.mdout',
                                 mutant_system.complex_prmtop, INPUT['surften'], csv_prefix,
                                 size, INPUT['dec_verbose'])
            pb_com.fill_all_terms()
            if INPUT['csv_format']:
                decompout.writerow(['Energy Decomposition Analysis (All units ' +
                                    'kcal/mol): Poisson Boltzmann Solvent (%s)' % mutstr])
                decompout.writerow([])
                pb_com.write_summary_csv(pb_com.numframes, decompout)
                decompout.writerow([])
            else:
                decompout.writeline('Energy Decomposition Analysis (All units ' +
                                    'kcal/mol): Generalized Born Solvent (%s)' % mutstr)
                decompout.writeline('')
                pb_com.write_summary(pb_com.numframes, decompout)
                decompout.writeline('')
            # Dump the energy vectors if requested
            if FILES.dec_energies:
                del pb_com.csvwriter
                dec_energies.write('PB Decomposition Energies (%s mutant)' %
                                   mutstr + ls)
                for token in gb_com.allowed_tokens:
                    utils.concatenate(dec_energies, csv_prefix + '.' + token + '.csv')
                dec_energies.write(ls)

    if FILES.dec_energies: dec_energies.close()
    dec_out_file.close()


# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

def write_decomp_binding_output(FILES, INPUT, size, prmtop_system,
                                mutant_system, mutstr, pre):
    """ Write output file for binding free energy decomposition calculations """
    from csv import writer
    from datetime import datetime
    from GMXMMPBSA.amber_outputs import (DecompOut, PairDecompOut, DecompBinding,
                               PairDecompBinding, MultiTrajDecompBinding,
                               MultiTrajPairDecompBinding)

    multitraj = bool(FILES.receptor_trajs or FILES.ligand_trajs)
    # Single trajectory
    if not multitraj:
        # Per-residue
        if INPUT['idecomp'] in [1, 2]:
            BindingClass = DecompBinding
            SingleClass = DecompOut
        # Pairwise
        else:
            BindingClass = PairDecompBinding
            SingleClass = PairDecompOut
    else:  # Multiple trajectories
        # Per-residue
        if INPUT['idecomp'] in [1, 2]:
            BindingClass = MultiTrajDecompBinding
            SingleClass = DecompOut
        # Pairwise
        else:
            BindingClass = MultiTrajPairDecompBinding
            SingleClass = PairDecompOut

    # First open up our output file and turn it into a CSV writer if necessary
    if INPUT['csv_format']:
        dec_out_file = OutputFile(FILES.decompout, 'wb', 0)
        decompout = writer(dec_out_file)
        decompout.writerow(['| Run on %s' % datetime.now().ctime()])
        if INPUT['gbrun']:
            decompout.writerow(['| GB non-polar solvation energies calculated '
                                'with gbsa=2'])
    else:
        dec_out_file = OutputFile(FILES.decompout, 'w', 0)
        decompout = dec_out_file
        decompout.write_date()
        if INPUT['gbrun']:
            decompout.add_comment('| GB non-polar solvation energies calculated '
                                  'with gbsa=2')

    # Open up the CSV energy vector file
    if FILES.dec_energies: dec_energies = open(FILES.dec_energies, 'w')

    # First we do GB
    if INPUT['gbrun']:
        # Setup for energy dump
        if FILES.dec_energies:
            csv_prefix = pre + 'gb_%s'
        else:
            csv_prefix = csv_pre = ''
        # Normal system
        if not INPUT['mutant_only']:
            if csv_prefix: csv_pre = csv_prefix % 'com'
            gb_com = SingleClass(pre + 'complex_gb.mdout',
                                 prmtop_system.complex_prmtop, INPUT['surften'],
                                 csv_pre, size, INPUT['dec_verbose'])
            if csv_prefix: csv_pre = csv_prefix % 'rec'
            gb_rec = SingleClass(pre + 'receptor_gb.mdout',
                                 prmtop_system.receptor_prmtop, INPUT['surften'],
                                 csv_pre, size, INPUT['dec_verbose'])
            if csv_prefix: csv_pre = csv_prefix % 'lig'
            gb_lig = SingleClass(pre + 'ligand_gb.mdout',
                                 prmtop_system.ligand_prmtop, INPUT['surften'],
                                 csv_pre, size, INPUT['dec_verbose'])
            if csv_prefix: csv_pre = pre + 'gb_bind'
            gb_bind = BindingClass(gb_com, gb_rec, gb_lig, prmtop_system,
                                   INPUT['idecomp'], INPUT['dec_verbose'], decompout,
                                   csv_pre, 'Energy Decomposition Analysis ' +
                                   '(All units kcal/mol): Generalized Born solvent')
            # Write the data to the output file
            gb_bind.parse_all()
            # Now it's time to dump everything to the CSV file
            if FILES.dec_energies:
                del gb_com.csvwriter, gb_rec.csvwriter, gb_lig.csvwriter
                del gb_bind.csvwriter
                dec_energies.write('Generalized Born Decomposition Energies' + ls)
                dec_energies.write(ls + 'Complex:' + ls)
                for token in gb_com.allowed_tokens:
                    utils.concatenate(dec_energies, pre + 'gb_com.' + token + '.csv')
                dec_energies.write(ls + 'Receptor:' + ls)
                for token in gb_rec.allowed_tokens:
                    utils.concatenate(dec_energies, pre + 'gb_rec.' + token + '.csv')
                dec_energies.write(ls + 'Ligand:' + ls)
                for token in gb_lig.allowed_tokens:
                    utils.concatenate(dec_energies, pre + 'gb_lig.' + token + '.csv')
                # DELTAs for each frame only computed if this is a single trajectory
                if not multitraj:
                    dec_energies.write(ls + 'DELTAS:' + ls)
                    for token in gb_com.allowed_tokens:
                        utils.concatenate(dec_energies, pre + 'gb_bind.' + token + '.csv')
                    dec_energies.write(ls)
        # Mutant system
        if INPUT['alarun']:
            if csv_prefix: csv_pre = csv_prefix % 'com'
            gb_com = SingleClass(pre + 'mutant_complex_gb.mdout',
                                 mutant_system.complex_prmtop, INPUT['surften'],
                                 csv_pre, size, INPUT['dec_verbose'])
            if csv_prefix: csv_pre = csv_prefix % 'rec'
            gb_rec = SingleClass(pre + 'mutant_receptor_gb.mdout',
                                 mutant_system.receptor_prmtop, INPUT['surften'],
                                 csv_pre, size, INPUT['dec_verbose'])
            if csv_prefix: csv_pre = csv_prefix % 'lig'
            gb_lig = SingleClass(pre + 'mutant_ligand_gb.mdout',
                                 mutant_system.ligand_prmtop, INPUT['surften'],
                                 csv_pre, size, INPUT['dec_verbose'])
            if csv_prefix: csv_pre = pre + 'gb_bind'
            gb_bind = BindingClass(gb_com, gb_rec, gb_lig, mutant_system,
                                   INPUT['idecomp'], INPUT['dec_verbose'], decompout,
                                   csv_pre, 'Energy Decomposition Analysis ' +
                                   '(All units kcal/mol): Generalized Born solvent (%s)' %
                                   mutstr)
            # Write the data to the output file
            gb_bind.parse_all()
            # Now it's time to dump everything to the CSV file
            if FILES.dec_energies:
                del gb_com.csvwriter, gb_rec.csvwriter, gb_lig.csvwriter
                del gb_bind.csvwriter
                dec_energies.write('GB Decomposition Energies (%s mutant)' %
                                   mutstr + ls)
                dec_energies.write(ls + 'Complex:' + ls)
                for token in gb_com.allowed_tokens:
                    utils.concatenate(dec_energies, pre + 'gb_com.' + token + '.csv')
                dec_energies.write(ls + 'Receptor:' + ls)
                for token in gb_rec.allowed_tokens:
                    utils.concatenate(dec_energies, pre + 'gb_rec.' + token + '.csv')
                dec_energies.write(ls + 'Ligand:' + ls)
                for token in gb_lig.allowed_tokens:
                    utils.concatenate(dec_energies, pre + 'gb_lig.' + token + '.csv')
                # DELTAs for each frame only computed if this is a single trajectory
                if not multitraj:
                    dec_energies.write(ls + 'DELTAS:' + ls)
                    for token in gb_com.allowed_tokens:
                        utils.concatenate(dec_energies, pre + 'gb_bind.' + token + '.csv')
                    dec_energies.write(ls)
    # Next do PB
    if INPUT['pbrun']:
        # Setup for energy dump
        if FILES.dec_energies:
            csv_prefix = pre + 'pb_%s'
        else:
            csv_prefix = csv_pre = ''
        # Normal system
        if not INPUT['mutant_only']:
            if csv_prefix: csv_pre = csv_prefix % 'com'
            pb_com = SingleClass(pre + 'complex_pb.mdout',
                                 prmtop_system.complex_prmtop, INPUT['cavity_surften'],
                                 csv_pre, size, INPUT['dec_verbose'])
            if csv_prefix: csv_pre = csv_prefix % 'rec'
            pb_rec = SingleClass(pre + 'receptor_pb.mdout',
                                 prmtop_system.receptor_prmtop, INPUT['cavity_surften'],
                                 csv_pre, size, INPUT['dec_verbose'])
            if csv_prefix: csv_pre = csv_prefix % 'lig'
            pb_lig = SingleClass(pre + 'ligand_pb.mdout',
                                 prmtop_system.ligand_prmtop, INPUT['cavity_surften'],
                                 csv_pre, size, INPUT['dec_verbose'])
            if csv_prefix: csv_pre = pre + 'pb_bind'
            pb_bind = BindingClass(pb_com, pb_rec, pb_lig, prmtop_system,
                                   INPUT['idecomp'], INPUT['dec_verbose'], decompout,
                                   csv_pre, 'Energy Decomposition Analysis ' +
                                   '(All units kcal/mol): Poisson Boltzmann solvent')
            # Write the data to the output file
            pb_bind.parse_all()
            # Now it's time to dump everything to the CSV file
            if FILES.dec_energies:
                del pb_com.csvwriter, pb_rec.csvwriter, pb_lig.csvwriter
                del pb_bind.csvwriter
                dec_energies.write('Poisson Boltzmann Decomposition Energies' + ls)
                dec_energies.write(ls + 'Complex:' + ls)
                for token in pb_com.allowed_tokens:
                    utils.concatenate(dec_energies, pre + 'pb_com.' + token + '.csv')
                dec_energies.write(ls + 'Receptor:' + ls)
                for token in pb_rec.allowed_tokens:
                    utils.concatenate(dec_energies, pre + 'pb_rec.' + token + '.csv')
                dec_energies.write(ls + 'Ligand:' + ls)
                for token in pb_lig.allowed_tokens:
                    utils.concatenate(dec_energies, pre + 'pb_lig.' + token + '.csv')
                # DELTAs for each frame only computed if this is a single trajectory
                if not multitraj:
                    dec_energies.write(ls + 'DELTAS:' + ls)
                    for token in pb_com.allowed_tokens:
                        utils.concatenate(dec_energies, pre + 'pb_bind.' + token + '.csv')
                    dec_energies.write(ls)
        # Mutant system
        if INPUT['alarun']:
            if csv_prefix: csv_pre = csv_prefix % 'com'
            pb_com = SingleClass(pre + 'mutant_complex_pb.mdout',
                                 mutant_system.complex_prmtop, INPUT['cavity_surften'],
                                 csv_pre, size, INPUT['dec_verbose'])
            if csv_prefix: csv_pre = csv_prefix % 'rec'
            pb_rec = SingleClass(pre + 'mutant_receptor_pb.mdout',
                                 mutant_system.receptor_prmtop, INPUT['cavity_surften'],
                                 csv_pre, size, INPUT['dec_verbose'])
            if csv_prefix: csv_pre = csv_prefix % 'lig'
            pb_lig = SingleClass(pre + 'mutant_ligand_pb.mdout',
                                 mutant_system.ligand_prmtop, INPUT['cavity_surften'],
                                 csv_pre, size, INPUT['dec_verbose'])
            if csv_prefix: csv_pre = pre + 'pb_bind'
            pb_bind = BindingClass(pb_com, pb_rec, pb_lig, mutant_system,
                                   INPUT['idecomp'], INPUT['dec_verbose'], decompout,
                                   csv_pre, 'Energy Decomposition Analysis ' +
                                   '(All units kcal/mol): Poisson Boltzmann solvent (%s)' %
                                   mutstr)
            # Write the data to the output file
            pb_bind.parse_all()
            if FILES.dec_energies:
                del pb_com.csvwriter, pb_rec.csvwriter, pb_lig.csvwriter
                del pb_bind.csvwriter
                dec_energies.write('PB Decomposition Energies (%s mutant)' %
                                   mutstr + ls)
                dec_energies.write(ls + 'Complex:' + ls)
                for token in pb_com.allowed_tokens:
                    utils.concatenate(dec_energies, pre + 'pb_com.' + token + '.csv')
                dec_energies.write(ls + 'Receptor:' + ls)
                for token in pb_rec.allowed_tokens:
                    utils.concatenate(dec_energies, pre + 'pb_rec.' + token + '.csv')
                dec_energies.write(ls + 'Ligand:' + ls)
                for token in pb_lig.allowed_tokens:
                    utils.concatenate(dec_energies, pre + 'pb_lig.' + token + '.csv')
                # DELTAs for each frame only computed if this is a single trajectory
                if not multitraj:
                    dec_energies.write(ls + 'DELTAS:' + ls)
                    for token in gb_com.allowed_tokens:
                        utils.concatenate(dec_energies, pre + 'pb_bind.' + token + '.csv')
                    dec_energies.write(ls)

    # Close the file(s)
    if FILES.dec_energies: dec_energies.close()
    dec_out_file.close()


# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class OutputFile(object):
    """ Main output file """

    # ==================================================

    def __init__(self, fname, mode='r', buffer=0):
        # Buffer is ignored
        self._binary = 'b' in mode.lower()
        self._handle = open(fname, mode)

    # ==================================================

    def write(self, stuff):
        # Handle bytes/str division in Py3 transparently
        if self._binary:
            if isinstance(stuff, str):
                self._handle.write(stuff.encode())
            else:
                self._handle.write(stuff)
        else:
            if isinstance(stuff, str):
                self._handle.write(stuff)
            else:
                self._handle.write(stuff.decode())
        self._handle.flush()

    # ==================================================

    def separate(self):
        """ Delimiter between fields in output file """
        for i in range(2):
            self.write('-' * 79)
            self.write(ls)

    # ==================================================

    def add_section(self, input_str):
        """ Adds a section to the output file """
        self.write(input_str)
        self.separate()

    # ==================================================

    def write_date(self):
        """ Writes a date string to the output file """
        from datetime import datetime
        self.writeline('| Run on %s' % datetime.now().ctime())

    # ==================================================

    def print_file_info(self, FILES, INPUT):
        """ Prints the summary information to a file """
        import GMXMMPBSA
        stability = not FILES.receptor_prmtop

        self.writeline('|gmx_MMPBSA Version=%s based on MMPBSA.py v.%s' % (GMXMMPBSA.__version__,
                                                                              GMXMMPBSA.__mmpbsa_version__))

        # if FILES.solvated_prmtop:
        #    self.writeline('|Solvated complex topology file:  %s' %
        #               FILES.solvated_prmtop)
        self.writeline('|Complex topology file:           %s' %
                       FILES.complex_prmtop)

        if not stability:
            # if FILES.receptor_trajs:
            #    if FILES.solvated_receptor_prmtop:
            #       self.writeline('|Solvated receptor topology file: %s' %
            #                  FILES.solvated_receptor_prmtop)
            self.writeline('|Receptor topology file:          %s' %
                           FILES.receptor_prmtop)

            # if FILES.ligand_trajs:
            # if FILES.solvated_ligand_prmtop:
            #    self.writeline('|Solvated ligand topology file:   %s' %
            #               FILES.solvated_ligand_prmtop)
            self.writeline('|Ligand topology file:            %s' %
                           FILES.ligand_prmtop)

        if INPUT['alarun']:
            self.writeline('|Mutant complex topology file:    %s' %
                           FILES.mutant_complex_prmtop)
            if not stability:
                self.writeline('|Mutant receptor topology file:   %s' %
                               FILES.mutant_receptor_prmtop)
                self.writeline('|Mutant ligand topology file:     %s' %
                               FILES.mutant_ligand_prmtop)

        self.write('|Initial mdcrd(s):                ')
        for i in range(len(FILES.complex_trajs)):
            if i == 0:
                self.writeline(FILES.complex_trajs[i])
            else:
                self.writeline('|                                 %s' %
                               FILES.complex_trajs[i])

        if FILES.receptor_trajs:
            self.write('|Initial Receptor mdcrd(s):       ')
            for i in range(len(FILES.receptor_trajs)):
                if i == 0:
                    self.writeline(FILES.receptor_trajs[i])
                else:
                    self.writeline('|                                 %s' %
                                   FILES.receptor_trajs[i])

        if FILES.ligand_trajs:
            self.write('|Initial Ligand mdcrd(s):         ')
            for i in range(len(FILES.ligand_trajs)):
                if i == 0:
                    self.writeline(FILES.ligand_trajs[i])
                else:
                    self.writeline('|                                 %s' %
                                   FILES.ligand_trajs[i])

    # ==================================================

    def add_comment(self, comment):
        """ Adds a comment string to the output file """
        self.writeline('|' + comment)

    # ==================================================

    def writeline(self, line):
        """ Adds a newline to the end of the passed line and writes it """
        self.write(line + ls)

    # ==================================================

    def __getattr__(self, attr):
        return getattr(self._handle, attr)

    # ==================================================

    def __del__(self):
        try:
            self._handle.close()
        except:
            pass

# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+



"""
This module contains classes and functions that manipulate Amber prmtop
objects with respect to MM/PBSA. It will figure out where atoms belong
in the context of a complex/receptor/ligand system. It will also be able
to translate complex atom selections into either Amber masks or Amber
group input selection strings for the complex, receptor, and ligand.
It will also check through the systems and make sure that the prmtop
files are compatible, as well. Necessary for gmx_MMPBSA functioning.
"""

# ##############################################################################
#                           GPLv3 LICENSE INFO                                 #
#                                                                              #
#  Copyright (C) 2020  Mario S. Valdes-Tresanco and Mario E. Valdes-Tresanco   #
#  Copyright (C) 2014  Jason Swails, Bill Miller III, and Dwight McGee         #
#                                                                              #
#   Project: https://github.com/Valdes-Tresanco-MS/gmx_MMPBSA                  #
#                                                                              #
#   This program is free software; you can redistribute it and/or modify it    #
#  under the terms of the GNU General Public License version 3 as published    #
#  by the Free Software Foundation.                                            #
#                                                                              #
#  This program is distributed in the hope that it will be useful, but         #
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
#  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    #
#  for more details.                                                           #
# ##############################################################################

from parmed.amber import LoadParm, AmberMask
from GMXMMPBSA.exceptions import PrmtopError, SelectionError


# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class Residue(object):
    """ Atom class in MM/PBSA, complex/receptor/ligand context """

    # -#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def __init__(self, number, residue):
        """ Initializes an Atom instance. NOTE: Atom numbers here index
          beginning at 1 for both number (complex) and receptor_number
          or ligand_number! Normal python arrays index from 0
      """
        self.number = number
        self.residue = residue
        self.receptor_number = None
        self.ligand_number = None
        self.selected = False

    # -#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def select(self):
        """ Label this atom as selected """
        self.selected = True

    # -#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def deselect(self):
        """ Deselect this atom """
        self.selected = False


# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class MMPBSA_System(object):
    """
    Sets up a system of a complex, receptor, and ligand.

    Get amber topology file from GROMACS

    The prmtops get initialized to prmtop objects immediately, so they must exist.
    Use it as:

    new_sys = MMPBSA_system(complex_prmtop, receptor_prmtop, ligand_prmtop)
    new_sys.Map()

    The MMPBSA_system is now set up, and you can get a Mask or Group input
    string. You can also check charges

    The following will return an amber mask string of the selected residues
    in the comma-delimited list format assuming you want a mask that corresponds
    to residue numbers in the complex prmtop (by definition, selection MUST
    correspond to residue numbers in the complex prmtop)

    com_mask, rec_mask, lig_mask = new_sys.Mask(selection, in_complex=False)

    The following will return a GROUP input string according to the same
    criteria as above for the masks

    com_group, rec_group, lig_group = new_sys.Group(selection, in_complex=True)

    The following will run sanity checks on the gmx_MMPBSA system to make sure stuff
    is consistent.

    new_sys.CheckConsistency()
    """

    # -#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def __init__(self, complex_prmtop, receptor_prmtop=None, ligand_prmtop=None):
        """
         Get amber topology from gromacs tpr file and Initializes the topology file classes for all of the prmtop files
         """

        self.complex_prmtop = LoadParm(complex_prmtop)
        self.stability = not receptor_prmtop and not ligand_prmtop
        if not self.stability:
            self.receptor_prmtop = LoadParm(receptor_prmtop)
            self.ligand_prmtop = LoadParm(ligand_prmtop)
        else:
            self.receptor_prmtop = None
            self.ligand_prmtop = None
        self.mapped = False
        self.ligstart = -1

        self._validate()

    # -#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def Map(self, receptor_mask=None, ligand_mask=None):
        """
        Gets the list of residues from the amber masks it receives
        :param receptor_mask: receptor amber mask
        :param ligand_mask: ligand amber mask
        :return:
        """

        # Begin find our own masks!
        if self.stability:
            # Stability mapping -- just copy the complex residue sequence
            self.res_list = \
                [Residue(i + 1, self.complex_prmtop.parm_data['RESIDUE_LABEL'][i])
                 for i in range(self.complex_prmtop.ptr('nres'))]
            self.mapped = True
            return

        rmask = AmberMask(self.complex_prmtop, receptor_mask).Selection()
        lmask = AmberMask(self.complex_prmtop, ligand_mask).Selection()
        # Check that every atom is selected once and only once by comparing
        # the sums of the two mask selections
        if sum(rmask) + sum(lmask) != self.complex_prmtop.ptr('natom'):
            raise PrmtopError("provided receptor/ligand masks don't select " +
                              "every atom in the complex topology!")
        if sum(rmask) != self.receptor_prmtop.ptr('natom'):
            raise PrmtopError('mismatch in receptor mask and receptor prmtop')
        if sum(lmask) != self.ligand_prmtop.ptr('natom'):
            raise PrmtopError('mismatch in ligand mask and ligand prmtop')

        # Now check that the masks don't select an atom twice
        for i in range(len(rmask)):
            if rmask[i] == 1 and lmask[i] == 1:
                raise PrmtopError('Atom %d selected by both receptor and ' % i +
                                  'ligand masks')
            if lmask[i] == 0 and rmask[i] == 0:
                raise PrmtopError('Atom %d is not selected by either ' % i +
                                  'receptor or ligand masks')

        # Now that we've verified everything is OK (we could do more, but this
        # should be good enough), let's actually create the res_list mapping
        lignum = recnum = 1
        self.res_list = []
        for i in range(self.complex_prmtop.ptr('nres')):
            start_ptr = self.complex_prmtop.parm_data['RESIDUE_POINTER'][i] - 1
            in_lig = lmask[start_ptr] == 1
            new_res = Residue(i + 1, self.complex_prmtop.parm_data['RESIDUE_LABEL'][i])
            if in_lig:
                new_res.ligand_number = lignum
                if (self.complex_prmtop.parm_data['RESIDUE_LABEL'][i] !=
                        self.ligand_prmtop.parm_data['RESIDUE_LABEL'][lignum - 1]):
                    raise PrmtopError('Residue mismatch while mapping. ' +
                                      'Incompatible topology files or bad mask definitions')
                lignum += 1
            else:
                new_res.receptor_number = recnum
                if (self.complex_prmtop.parm_data['RESIDUE_LABEL'][i] !=
                        self.receptor_prmtop.parm_data['RESIDUE_LABEL'][recnum - 1]):
                    raise PrmtopError('Residue mismatch while mapping. ' +
                                      'Incompatible topology files or bad mask definitions')
                recnum += 1
            self.res_list.append(new_res)
        # end for i in range(self.complex_prmtop.ptr('nres'))

        self.mapped = True
        return  # done here

    # -#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def Mask(self, selection, in_complex=True):
        """ This function turns a ,- or ;-delimited list of residues into an
          Amber mask for the complex, receptor, and ligand. If in_complex is
          true, then the masks returned will correspond to the residues inside
          the complex. If it's false, then it will correspond to just the
          selection in the receptor/ligand prmtops
      """

        if selection.lower() == 'all':
            selection = '1-%d' % self.complex_prmtop.ptr('nres')

        if self.stability:
            # M.L. added check for single residue
            if self.complex_prmtop.ptr('nres') == 1:
                return (':1', None, None)
            else:
                return self._stability_mask(selection)
        else:
            return self._binding_mask(selection, in_complex)

    # -#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def _stability_mask(self, selection):
        """ Internal call for Mask if stability is true """

        self._deselect_residues()
        self._select_residues(selection)

        # M.L. added a first residue condition, so do not put comma before it
        first_residue = False
        com_start = 0
        mask = ':'
        # M.L. keeps i as 0-index, use resid as 1-index for printout for cpptraj
        i = 0
        resid = 1
        while i < self.complex_prmtop.ptr('nres'):

            if not self.res_list[i].selected:
                i += 1
                resid += 1
                continue

            else:
                com_start = resid
                # M.L. modified loop
                # check that the next residue j is in the selection
                while i < self.complex_prmtop.ptr('nres') - 1:  # nres-1 b/c searching for j
                    j = i + 1
                    if self.res_list[j].selected:
                        i += 1
                        resid += 1
                    else:
                        break
                if first_residue == False:
                    mask += range_string(com_start, resid)
                    first_residue = True
                else:
                    mask += ',' + range_string(com_start, resid)
                com_start = -1
                i += 1
                resid += 1

        return (mask, None, None)

    # -#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def _binding_mask(self, selection, in_complex):
        """ Internal call if stability is False """

        self._deselect_residues()
        self._select_residues(selection)
        com_start = -1
        rec_start = -1
        lig_start = -1
        rec_end = -1
        lig_end = -1
        complex_mask = ''
        receptor_mask = ''
        ligand_mask = ''

        # Create 0ed arrays for receptor and ligand selections. If in_complex is
        # False, then we will need these arrays to generate the masks, or we run
        # the risk of artificially fragmented masks

        receptor_selection = [0 for i in range(self.receptor_prmtop.ptr('nres'))]
        ligand_selection = [0 for i in range(self.ligand_prmtop.ptr('nres'))]

        # Now every residue in self.res_list has been selected according to
        # that selection. Now we have to loop through all of the atoms until
        # we hit one that's selected
        i = 0
        while i < len(self.res_list):

            # If we have any intermediate receptor/ligand selections, flush them
            # here. i has been added twice too many times, so we need to subtract
            # off 2 of them, but add 1 back to adjust from 0 -> 1 indexing
            if rec_start != -1:
                if rec_start == rec_end:
                    receptor_mask += ',%d' % rec_start
                else:
                    receptor_mask += ',%d-%d' % (rec_start, i - 1)
                rec_start = -1
            if lig_start != -1:
                if lig_start == lig_end:
                    ligand_mask += ',%d' % lig_start
                else:
                    ligand_mask += ',%d-%d' % (lig_start, i - 1)
                lig_start = -1

            # Skip over unselected atoms
            if not self.res_list[i].selected:
                i += 1
                continue

            # Now this atom is selected. Add it to where it belongs
            com_start = i + 1
            if self.res_list[i].receptor_number != None:  # it's in receptor
                if in_complex:
                    rec_start = i + 1
                else:
                    receptor_selection[self.res_list[i].receptor_number - 1] = 1
                in_rec = True
            elif self.res_list[i].ligand_number != None:  # it's in ligand
                if in_complex:
                    lig_start = i + 1
                else:
                    ligand_selection[self.res_list[i].ligand_number - 1] = 1
                in_rec = False

            i += 1

            while i < len(self.res_list) and self.res_list[i].selected:
                # Now we have to make sure that we didn't leave the receptor
                # or ligand, or update where we are if we did.
                if in_rec:
                    if self.res_list[i].receptor_number != None:
                        # We're still in the receptor, go on
                        receptor_selection[self.res_list[i].receptor_number - 1] = 1
                        i += 1
                        continue
                    else:
                        # Define where we end and add to mask if in_complex
                        if in_complex:
                            rec_end = i
                            receptor_mask += ',' + range_string(rec_start, rec_end)
                            rec_start = -1
                        in_rec = False
                        # Define the start of our ligand
                        ligand_selection[self.res_list[i].ligand_number - 1] = 1
                        if in_complex: lig_start = i + 1
                        i += 1
                        continue
                else:
                    if self.res_list[i].ligand_number != None:
                        # We're still in the ligand, go on
                        ligand_selection[self.res_list[i].ligand_number - 1] = 1
                        i += 1
                        continue
                    else:
                        # Define where we end and add to mask if in_complex
                        if in_complex:
                            lig_end = i
                            ligand_mask += ',' + range_string(lig_start, lig_end)
                            lig_start = -1  # so we know this one is ended
                        in_rec = True
                        # Define the start of our receptor
                        receptor_selection[self.res_list[i].receptor_number - 1] = 1
                        if in_complex: rec_start = i + 1
                        i += 1
                        continue

                raise SelectionError("About to enter an infinite loop. Oh no!")

                # end if in_rec

            # end while i ... selected

            com_end = i
            complex_mask += ',' + range_string(com_start, com_end)
            com_start = -1

            i += 1

        # end while i < len(self.res_list)

        # If we had selected the last residue, then we have to add that last
        # chunk of mask back in.
        if com_start != -1:
            complex_mask += ',' + range_string(com_start, len(self.res_list))

        # Now we have a choice in how to construct our masks. If in_complex, then
        # just use them as they've been created and replace the beginning , with :
        # If not in_complex, then we need to use receptor_selection and
        # ligand_selection to determine the masks.

        if in_complex:
            if rec_start != -1:
                if in_complex:
                    rec_end = len(self.res_list)
                else:
                    rec_end = self.res_list[len(self.res_list) - 1].receptor_number
                receptor_mask += ',' + range_string(rec_start, rec_end)

            if lig_start != -1:
                if in_complex:
                    lig_end = len(self.res_list)
                else:
                    lig_end = self.res_list[len(self.res_list) - 1].ligand_number
                ligand_mask += ',' + range_string(lig_start, lig_end)
        else:
            # Get receptor mask
            i = 0
            start = -1
            while i < len(receptor_selection):
                if receptor_selection[i] == 1:
                    start = i + 1
                    i += 1
                    while i < len(receptor_selection) and receptor_selection[i] == 1:
                        i += 1
                    if start == i:
                        receptor_mask += ',%d' % start
                    else:
                        receptor_mask += ',%d-%d' % (start, i)
                    start = -1
                i += 1

            if start != -1:
                # The end was selected
                if start == len(receptor_selection):
                    receptor_mask += ',%d' % start
                else:
                    receptor_mask += ',%d-%d' % (start, len(receptor_selection))

            # Get ligand mask
            i = 0
            start = -1
            while i < len(ligand_selection):
                if ligand_selection[i] == 1:
                    start = i + 1
                    i += 1
                    while i < len(ligand_selection) and ligand_selection[i] == 1:
                        i += 1
                    if start == i:
                        ligand_mask += ',%d' % start
                    else:
                        ligand_mask += ',%d-%d' % (start, i)
                    start = -1
                i += 1

            if start != -1:
                # The end was selected
                if start == len(ligand_selection):
                    ligand_mask += ',%d' % start
                else:
                    ligand_mask += ',%d-%d' % (start, len(ligand_selection))

        # end if in_complex

        # Now, for any existing masks, replace leading , with :

        if complex_mask != '':
            complex_mask = ':' + complex_mask[1:]
        if ligand_mask != '':
            ligand_mask = ':' + ligand_mask[1:]
        if receptor_mask != '':
            receptor_mask = ':' + receptor_mask[1:]

        return (complex_mask, receptor_mask, ligand_mask)

    # -#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def Group(self, selection, in_complex=True, com_group_type='RES',
              rec_group_type='RRES', lig_group_type='LRES'):
        """ This function turns a ,- or ;-delimited list of residues into an
          Amber group input specification for the complex, receptor, and ligand.
          If in_complex is true, then the group inputs returned will correspond
          to the residues inside the complex. If it's false, then it will
          correspond to just the selection in the receptor/ligand prmtops
        """

        if selection.lower() == 'all':
            selection = '1-%d' % self.complex_prmtop.ptr('nres')

        if self.stability:
            return self._stability_group(selection, com_group_type, rec_group_type)
        else:
            return self._binding_group(selection, in_complex, com_group_type,
                                       rec_group_type, lig_group_type)

    # -#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def _stability_group(self, selection, com_group_type, rec_group_type):
        """ Internal function for Group for stability calcs """

        self._deselect_residues()
        self._select_residues(selection)
        MAX_GROUP = 7
        com_start = -1

        i = 0
        num_groups = 1
        com_group_string = '%s' % com_group_type
        rec_group_string = '%s' % rec_group_type
        while i < self.complex_prmtop.ptr('nres'):
            if not self.res_list[i].selected:
                i += 1
                continue

            else:
                com_start = i + 1
                i += 1
                while i < self.complex_prmtop.ptr('nres') and \
                        self.res_list[i].selected:
                    i += 1

                if num_groups == MAX_GROUP:
                    com_group_string += '\n%s' % com_group_type
                    rec_group_string += '\n%s' % rec_group_type

                com_group_string += ' %d %d' % (com_start, i)
                rec_group_string += ' %d %d' % (com_start, i)
                num_groups += 1
                i += 1

        # end while i < len(....(nres))

        return (com_group_string, rec_group_string, '')

    # -#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def _binding_group(self, selection, in_complex, com_group_type,
                       rec_group_type, lig_group_type):
        """ Internal function for Group for no stability """
        import re
        numre = re.compile(r'(\d+)')

        self._deselect_residues()
        self._select_residues(selection)
        MAX_GROUP = 7
        com_start = -1
        rec_start = -1
        lig_start = -1
        complex_group = []
        receptor_group = []
        ligand_group = []

        # Create 0ed arrays for receptor and ligand selections. If in_complex is
        # False, then we will need these arrays to generate the masks, or we run
        # the risk of artificially fragmented masks

        receptor_selection = [0 for i in range(self.receptor_prmtop.ptr('nres'))]
        ligand_selection = [0 for i in range(self.ligand_prmtop.ptr('nres'))]

        # Now every residue in self.res_list has been selected according to
        # that selection. Now we have to loop through all of the atoms until
        # we hit one that's selected
        i = 0
        while i < len(self.res_list):

            # If we have any intermediate receptor/ligand selections, flush them
            # here. i has been added twice too many times, so we need to subtract
            # off 2 of them, but add 1 back to adjust from 0 -> 1 indexing
            if rec_start != -1:
                receptor_group.append('%d %d' % (rec_start, i - 1))
                rec_start = -1
            if lig_start != -1:
                ligand_group.append('%d %d' % (lig_start, i - 1))
                lig_start = -1

            # Skip over unselected atoms
            if not self.res_list[i].selected:
                i += 1
                continue

            # Now this atom is selected. Add it to where it belongs
            com_start = i + 1
            if self.res_list[i].receptor_number != None:  # it's in receptor
                if in_complex:
                    rec_start = i + 1
                else:
                    receptor_selection[self.res_list[i].receptor_number - 1] = 1
                in_rec = True
            elif self.res_list[i].ligand_number != None:  # it's in ligand
                if in_complex:
                    lig_start = i + 1
                else:
                    ligand_selection[self.res_list[i].ligand_number - 1] = 1
                in_rec = False

            i += 1

            while i < len(self.res_list) and self.res_list[i].selected:
                # Now we have to make sure that we didn't leave the receptor
                # or ligand, or update where we are if we did.
                if in_rec:
                    if self.res_list[i].receptor_number != None:
                        # We're still in the receptor, go on
                        receptor_selection[self.res_list[i].receptor_number - 1] = 1
                        i += 1
                        continue
                    else:
                        # Define where we end and add to mask if in_complex
                        if in_complex:
                            rec_end = i
                            receptor_group.append('%d %d' % (rec_start, rec_end))
                            rec_start = -1
                        in_rec = False
                        # Define the start of our ligand
                        ligand_selection[self.res_list[i].ligand_number - 1] = 1
                        if in_complex: lig_start = i + 1
                        i += 1
                        continue
                else:
                    if self.res_list[i].ligand_number != None:
                        # We're still in the ligand, go on
                        ligand_selection[self.res_list[i].ligand_number - 1] = 1
                        i += 1
                        continue
                    else:
                        # Define where we end and add to mask if in_complex
                        if in_complex:
                            lig_end = i
                            ligand_group.append('%d %d' % (lig_start, lig_end))
                            lig_start = -1  # so we know this one is ended
                        in_rec = True
                        # Define the start of our receptor
                        receptor_selection[self.res_list[i].receptor_number - 1] = 1
                        if in_complex: rec_start = i + 1
                        i += 1
                        continue

                raise SelectionError("About to enter an infinite loop. Oh no!")

                # end if in_rec

            # end while i ... selected

            com_end = i
            complex_group.append('%d %d' % (com_start, com_end))
            com_start = -1

            i += 1

        # end while i < len(self.res_list)

        # If we had selected the last residue, then we have to add that last
        # chunk of mask back in.
        if com_start != -1:
            complex_group.append('%d %d' % (com_start, len(self.res_list)))

        # Now we have a choice in how to construct our masks. If in_complex, then
        # just use them as they've been created and replace the beginning , with :
        # If not in_complex, then we need to use receptor_selection and
        # ligand_selection to determine the masks.

        if in_complex:
            if rec_start != -1:
                receptor_group.append('%d %d' % (rec_start, len(self.res_list)))

            if lig_start != -1:
                ligand_group.append('%d %d' % (lig_start, len(self.res_list)))
        else:
            # Get receptor group
            i = 0
            start = -1
            while i < len(receptor_selection):
                if receptor_selection[i] == 1:
                    start = i + 1
                    i += 1
                    while i < len(receptor_selection) and receptor_selection[i] == 1:
                        i += 1
                    receptor_group.append('%d %d' % (start, i))
                    start = -1
                i += 1

            if start != -1:
                # The end was selected
                receptor_group.append('%d %d' % (start, len(receptor_selection)))

            # Get ligand group
            i = 0
            start = -1
            while i < len(ligand_selection):
                if ligand_selection[i] == 1:
                    start = i + 1
                    i += 1
                    while i < len(ligand_selection) and ligand_selection[i] == 1:
                        i += 1
                    ligand_group.append('%d %d' % (start, i))
                    start = -1
                i += 1

            if start != -1:
                # The end was selected
                ligand_group.append('%d %d' % (start, len(ligand_selection)))

        # end if in_complex

        # Now, for any existing groups, format them for group input, making
        # sure to limit the number of groups to MAX_GROUP

        com_grp_str = com_group_type + ' '
        rec_grp_str = rec_group_type + ' '
        lig_grp_str = lig_group_type + ' '

        # First do the complex
        num_grps = 0
        for grp in complex_group:
            if num_grps % MAX_GROUP == 0 and num_grps != 0:
                com_grp_str += '\n%s ' % com_group_type
            com_grp_str += grp + ' '
            num_grps += 1

        # Next do the receptor
        num_grps = 0
        for grp in receptor_group:
            if num_grps % MAX_GROUP == 0 and num_grps != 0:
                rec_grp_str += '\n%s ' % rec_group_type
            rec_grp_str += grp + ' '
            num_grps += 1

        # Next do the ligand
        num_grps = 0
        for grp in ligand_group:
            if num_grps % MAX_GROUP == 0 and num_grps != 0:
                lig_grp_str += '\n%s ' % lig_group_type
            lig_grp_str += grp + ' '
            num_grps += 1

        # Groups are invalid if they are empty. So if either rec_grp_str or
        # lig_grp_str have no numbers, add a 1 1 to it
        if len(numre.findall(rec_grp_str)) == 0:
            rec_grp_str += "1 1"
        if len(numre.findall(lig_grp_str)) == 0:
            lig_grp_str += "1 1"
        return (com_grp_str, rec_grp_str, lig_grp_str)

    # -#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def _select_residues(self, selection):
        """ This routine takes a given selection, comma- or ;-delimited,
          and selects the appropriate residues from res_list.
      """

        if not self.mapped:
            raise PrmtopError('The MMPBSA_system must be mapped before you can '
                              'request masks or groups!')

        selection = selection.replace(';', ',')  # move to one kind of delimiter
        selection_fields = selection.split(',')

        # Get rid of any blank fields

        for field in selection_fields:

            field = field.strip()

            # Skip over blank fields
            if len(field) == 0:
                continue

            # If '-' is in the field, then it's a range
            if '-' in field:
                field_items = field.split('-')

                # Make sure the fields are properly formed
                if len(field_items) != 2:
                    raise SelectionError('Selection ranges must be "# - #"!')
                if len(field_items[0].strip()) == 0:
                    raise SelectionError('Selection ranges must be "# - #"!')
                if len(field_items[1].strip()) == 0:
                    raise SelectionError('Selection ranges must be "# - #"!')

                # Make sure the fields are composed of integers
                try:
                    res1 = int(field_items[0])
                    res2 = int(field_items[1])
                except:
                    raise SelectionError('Invalid selection! Integers expected.')

                # Now make sure that they're within the legal range and not stupid
                if res2 < res1 or res1 <= 0 or \
                        res2 > self.complex_prmtop.ptr('nres'):
                    raise SelectionError('Invalid selection. Illegal residue range')

                # Now go through and select them
                for i in range(res1 - 1, res2):
                    self.res_list[i].select()

            else:
                # Here the field is just a single residue
                try:
                    res1 = int(field)
                except:
                    raise SelectionError('Invalid selection! Integers expected.')

                # select that one residue

                self.res_list[res1 - 1].select()

            # End '-' in field

        # End for field...:

    # -#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def _deselect_residues(self):
        """ Deselect all of the residues in the system """

        if not self.mapped:
            raise PrmtopError('Cannot deselect residues before mapping!')

        for i in range(len(self.res_list)):
            self.res_list[i].deselect()

    # -#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def _validate(self):
        """ Validates the passed prmtops. Does a preliminary compatibility
          check
      """
        if self.stability: return None  # done checking here for stability calcs

        if self.complex_prmtop.ptr('natom') != self.receptor_prmtop.ptr('natom') \
                + self.ligand_prmtop.ptr('natom'):
            raise PrmtopError('Complex natom != receptor natom + ligand natom')

        if self.complex_prmtop.ptr('nres') != self.receptor_prmtop.ptr('nres') + \
                self.ligand_prmtop.ptr('nres'):
            raise PrmtopError('Complex nres != receptor nres + ligand nres')

        if (self.complex_prmtop.chamber or self.receptor_prmtop.chamber or
                self.ligand_prmtop.chamber):
            if (not self.complex_prmtop.chamber or not self.receptor_prmtop.chamber
                    or not self.ligand_prmtop.chamber):
                raise PrmtopError('Prmtops must be either ALL chamber ' +
                                  'prmtops or NONE')

    # -#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def CheckConsistency(self):
        """ Checks to make sure that the charge definitions are consistent across
          each prmtop file, but only if we know which atoms are which atoms in
          each prmtop
      """

        if self.stability: return None

        # Check that our radii sets are consistent
        try:
            com_radii = self.complex_prmtop.parm_data['RADIUS_SET'][0]
            rec_radii = self.receptor_prmtop.parm_data['RADIUS_SET'][0]
            lig_radii = self.ligand_prmtop.parm_data['RADIUS_SET'][0]
            if com_radii != rec_radii or com_radii != lig_radii:
                raise PrmtopError('Topology files have inconsistent RADIUS_SETs')
        except KeyError:
            raise PrmtopError('Topology files have no Implicit radii! You can ' +
                              'add implicit radii using xparmed.py or parmed.py')
        if not self.mapped:
            raise PrmtopError('Cannot check charges prior to mapping!')

        TINY = 0.0001

        # Loop through every residue and compare its charge value in the complex
        # to the corresponding atom in the ligand or receptor
        for i in range(self.complex_prmtop.ptr('natom')):
            resnum = self.complex_prmtop.atoms[i].residue.idx
            atom_in_res = (i + 1 -
                           self.complex_prmtop.parm_data['RESIDUE_POINTER'][resnum])
            com_chg = self.complex_prmtop.parm_data['CHARGE'][i]
            if self.res_list[resnum].ligand_number is not None:
                otherparm = self.ligand_prmtop
                resnum = self.res_list[resnum].ligand_number - 1
            else:
                otherparm = self.receptor_prmtop
                resnum = self.res_list[resnum].receptor_number - 1

            atnum = otherparm.parm_data['RESIDUE_POINTER'][resnum] - 1 + atom_in_res
            lig_or_rec_chg = otherparm.parm_data['CHARGE'][atnum]
            if abs(lig_or_rec_chg - com_chg) > TINY:
                raise PrmtopError('Inconsistent charge definition for atom %d!' %
                                  (i + 1))


# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

def range_string(num1, num2):
    """ This function returns a single number if they're equal, or a range if
       they're not
   """

    if num1 == num2:
        return str(num1)
    else:
        return '%d-%d' % (num1, num2)

# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
